// Copyright Daniel Wallin, David Abrahams 2005. Use, modification and
// distribution is subject to the Boost Software License, Version 1.0. (See
// accompanying file LICENSE_1_0.txt or copy at
// http://www.boost.org/LICENSE_1_0.txt)
//
//
// Modified by Roland Pihlakas (2007) 
//

#ifndef TAGGED_ARGUMENT_050328_HPP
#define TAGGED_ARGUMENT_050328_HPP

#include <boost/parameter/aux_/void.hpp>
#include <boost/parameter/aux_/arg_list.hpp>
#include <boost/detail/is_xxx.hpp>

namespace boost { namespace parameter { namespace aux {

// Holds a reference to an argument of type Arg associated with
// keyword Keyword
    
    
#pragma warning (push)								    //roland
#pragma warning (disable : 4512)						//roland - warning C4512: 'boost::parameter::aux::tagged_argument<Keyword,Arg>' : assignment operator could not be generated	    
    
    
template <class Keyword, class Arg>
struct tagged_argument
{
    typedef Keyword key_type;
    typedef Arg value_type;
    typedef Arg& reference;

    inline tagged_argument(reference x) : value(x) {}





	// roland start
	
	template <class Keyword2, class Arg2>   
	struct add_param
	{
		typedef arg_list<														
			tagged_argument<Keyword2, Arg2>
		  , arg_list<tagged_argument<Keyword, Arg> > 
		>		
		type;
	};        

	template <class Arg2>   
	struct add_param<Keyword, Arg2>
	{
		typedef arg_list<													//inline by roland
			tagged_argument<Keyword, Arg2>			 //Arg2
		  , empty_arg_list 
		>		
		type;
	};        
	
	
/*	
	template <class Keyword2>
	struct typeof
	{
		BOOST_STATIC_ASSERT(false);
	};
	
	template <>
	struct typeof <Keyword>
	{
		typedef Arg				type;
	};
*/	

	// roland end




    // Comma operator to compose argument list without using parameters<>.
    // Useful for argument lists with undetermined length.
#if 0																				//#if 0 by roland
    template <class Keyword2, class Arg2>
    arg_list<
        tagged_argument<Keyword, Arg>
      , arg_list<tagged_argument<Keyword2, Arg2> > 
    >
    operator,(tagged_argument<Keyword2, Arg2> x) const
    {
        return arg_list<
            tagged_argument<Keyword, Arg>
          , arg_list<tagged_argument<Keyword2, Arg2> > 
        >(
            *this
          , arg_list<tagged_argument<Keyword2, Arg2> >(x, empty_arg_list())
        );
    }
#else																		   //#else section by roland	- i think, if we push later arguments into the beginning of the arg_list then lets do it here too!
	template <class Keyword2, class Arg2>
    inline typename add_param<Keyword2, Arg2>::type							//inline and add_param by roland
    operator,(tagged_argument<Keyword2, Arg2> x) const
    {
        return add_param<Keyword2, Arg2>::type									//add param by roland
        (
            x
          , arg_list<tagged_argument<Keyword, Arg> >(*this, empty_arg_list())
        );
    }
#endif																		   //#endif by roland - roland end
    
    
    
    
    //roland start - if new param has same key as previous, then remove ("overwrite") the previous one. Prevents error C2593: 'operator [' is ambiguous
    
	template <class Arg2>													//if prev param was tagged_argument. called in the form (p, key = val)
    inline typename add_param<Keyword, Arg2>::type									//inline by roland
    operator,(tagged_argument<Keyword, Arg2> x) const						
    {
        return add_param<Keyword, Arg2>::type
        (
            x						 //NB
          , empty_arg_list()
        );
    }
    
    
	template <class Arg2>													//if prev param was tagged_argument. called in the form (key = val, p)
    inline arg_list<												//inline by roland
        tagged_argument<Keyword, Arg>			//Arg
      , empty_arg_list 
    >
    operator,(tagged_argument<Keyword, Arg2> const &x) //const				//we must remove const specifier from here to disambiquate , operator		
    {
        return arg_list<            
            tagged_argument<Keyword, Arg>			//Arg
          , empty_arg_list
        >(
			*this										//NB
          , empty_arg_list()
        );
    }
    
    
    
    
    
    

    
	/*
    template <class Arg2, class arg_list3_T>								//if prev param was arg_list
    arg_list<
        tagged_argument<Keyword, Arg>
      , arg_list3_T 
    >
    operator,(arg_list<tagged_argument<Keyword, Arg2>, arg_list3_T> x) const	   //roland - existing argument list's head contains the same key as new one
    {
        return arg_list<
            tagged_argument<Keyword, Arg>
          , arg_list3_T 
        >(
            *this
          , static_cast<arg_list3_T>(x)
        );
    }
    */
   
    //roland end - if new param has same key as previous, then remove ("overwrite") the previous one. Prevents error C2593: 'operator [' is ambiguous
    
    
    
    
    
    //roland start - allows merging argument lists
    
    template <class arg_list2_T>
    /*arg_list<
        tagged_argument<Keyword, Arg>
      , arg_list2_T 
    >  */
    //typename arg_list2_T::merge_result_type<tagged_argument<Keyword, Arg2> >::type			//we dont have simple way to know merge result here, because whole list might get rebuilt
	//arg_list<tagged_argument<Keyword, Arg>, typename arg_list2_T::remove_key_from_index_T<arg_list2_T::index_of_key<Keyword>::rval>::type>
	inline															 //inline by roland
	typename arg_list2_T::replace_key_from_index_T<arg_list2_T::index_of_key<Keyword>::rval + 1, tagged_argument<Keyword, Arg> >::type		  //+1 is neccessary because MSVC release build for some reason does not match template of -1, it sees instead 0xFFFFFFFF
    operator,(arg_list2_T const &x) const
    {
		/*
        return arg_list<
            tagged_argument<Keyword, Arg>
          , arg_list2_T 
        >(
            *this
          , x
        );
        */
        return (x, *this);			//lets use arg_list's merging method
    }
        
        
        
    inline arg_list<												 //inline by roland
        tagged_argument<Keyword, Arg>			
      , empty_arg_list 
    >
    operator,(empty_arg_list const &x) const						//if previous argument list was empty		
    {																   	//this method is also useful for case when some macro sometimes adds a parameter and sometimes not (using then empty_arg_list instead), but we still want to use comma before that macro
        return arg_list<            
            tagged_argument<Keyword, Arg>		
          , empty_arg_list
        >(
			*this					
          , empty_arg_list()
        );
    }
        
        
        
        
        
    //roland end - allows merging argument lists
    
    
    
    
    
    
    //roland start - isset, unset and set methods
    
    template <class key>
	bool isset2(key &) const
	{
		return false;
	}
	
	template <>
	bool isset2<keyword<key_type> >(keyword<key_type> &) const
	{
		return true;
	}
	
	
	
	template <class key>
	struct isset
	{
		static const bool rval = false;
	};
	
	template <>
	struct isset<key_type>
	{
		static const bool rval = true;
	};

	
	
	
	
	
	template <class key>													 ///prevents Error C2785 'declaration1' and 'declaration2' have different return types for unset method
	struct get_unset_rT
	{
		typedef arg_list<tagged_argument<Keyword, Arg>, empty_arg_list>		type;		
	};
	
	template <>
	struct get_unset_rT<keyword<key_type> >
	{
		typedef empty_arg_list												type;
	};
	
	
	
    
    template <class key>
	typename get_unset_rT<key>::type
	unset(key &) const
	{
		return arg_list<            
            tagged_argument<Keyword, Arg>		
          , empty_arg_list
        >(
			*this					
          , empty_arg_list()
        );
	};
		
	//struct dummy_P{};
	
	
	template <>
	typename get_unset_rT<keyword<key_type> >::type
	unset<keyword<key_type> >(keyword<key_type> &) const
	{
		return empty_arg_list();
	}
	
	/*
	template <>
	typename get_unset_rT<key_type>::type
	unset<key_type>() const
	{
		return empty_arg_list();
	};
	*/
    
    
    //roland end - isset, unset and set methods
    
    
    
    
    
    
    
    
    //roland start - lets add support for single-keyword calls


	// A metafunction class that, given a keyword and a default
    // type, returns the appropriate result type for a keyword
    // lookup given that default
    struct binding
    {
        template <class KW, class Default>
        struct apply
        {
          typedef typename mpl::eval_if<
                boost::is_same<KW, key_type>
              , mpl::identity<reference>
              , mpl::apply_wrap2<typename empty_arg_list::binding, KW, Default>
          >::type type;
        };
    };




    //roland2 start - useful when we want to read some typedefs in argument's class
    
    struct binding_noref
    {
        template <class KW, class Default>
        struct apply
        {
          typedef typename mpl::eval_if<
                boost::is_same<KW, key_type>
              , mpl::identity<value_type>
              , mpl::apply_wrap2<typename empty_arg_list::binding_noref, KW, Default>
          >::type type;
        };
    };
    
    //roland2 end - useful when we want to read some typedefs in argument's class




    
#if BOOST_WORKAROUND(BOOST_MSVC, <= 1300) \
    || BOOST_WORKAROUND(__GNUC__, < 3) \
    || BOOST_WORKAROUND(__BORLANDC__, BOOST_TESTED_AT(0x564))
    // These older compilers don't support the overload set creation
    // idiom well, so we need to do all the return type calculation
    // for the compiler and dispatch through an outer function template




    // Outer indexing operators that dispatch to the right node's
    // get() function.
    template <class KW>
    typename mpl::apply_wrap2<binding, KW, void_>::type
    operator[](keyword<KW> const& x) const
    {
        typename mpl::apply_wrap1<key_owner, KW>::type const& sublist = *arg_list<tagged_argument<Keyword, Arg> >(*this, empty_arg_list());   //*this
        return sublist.get(x);
    }

    template <class KW, class Default>
    typename mpl::apply_wrap2<binding, KW, Default&>::type
    operator[](default_<KW, Default> x) const
    {
        typename mpl::apply_wrap1<key_owner, KW>::type const& sublist = *arg_list<tagged_argument<Keyword, Arg> >(*this, empty_arg_list());   //*this
        return sublist.get(x);
    }

    template <class KW, class F>
    typename mpl::apply_wrap2<
        binding,KW
      , typename result_of0<F>::type
    >::type
    operator[](lazy_default<KW,F> x) const
    {
        typename mpl::apply_wrap1<key_owner, KW>::type const& sublist = *arg_list<tagged_argument<Keyword, Arg> >(*this, empty_arg_list());   //*this
        return sublist.get(x);
    }

#else

    reference operator[](keyword<key_type> const& key) const
    {
        //return arg.value;
        return arg_list<tagged_argument<Keyword, Arg> >(*this, empty_arg_list())[key];
    }

    template <class Default>
    reference operator[](default_<key_type, Default> key) const
    {
		//return arg.value;
        return arg_list<tagged_argument<Keyword, Arg> >(*this, empty_arg_list())[key];
    }

    template <class Default>
    reference operator[](lazy_default<key_type, Default> key) const
    {					
		//return arg.value;
        return arg_list<tagged_argument<Keyword, Arg> >(*this, empty_arg_list())[key];
    }

    
    
    // If this function is called, it means there is no argument
    // in the list that matches the supplied keyword. Just return
    // the default value.
    template <class K, class Default>
    Default& operator[](default_<K, Default> x) const
    {
        return x.value;
    }

    // If this function is called, it means there is no argument
    // in the list that matches the supplied keyword. Just evaluate
    // and return the default value.
    template <class K, class F>
    typename result_of0<F>::type
    operator[](
        BOOST_PARAMETER_lazy_default_fallback<K,F> x) const
    {
        return x.compute_default();
    }
    
    
#endif    
    
    
    //roland end - lets add support for single-keyword calls
    
    
    
    
    
    reference value;
};

#pragma warning (pop)									//roland - warning C4512: 'boost::parameter::aux::tagged_argument<Keyword,Arg>' : assignment operator could not be generated	    
   


// Defines a metafunction, is_tagged_argument, that identifies
// tagged_argument specializations.
BOOST_DETAIL_IS_XXX_DEF(tagged_argument,tagged_argument,2)

}}} // namespace boost::parameter::aux

#endif // TAGGED_ARGUMENT_050328_HPP

